#include <gsl/gsl_sf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "math.h"
#include <gst/gst.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>
#include "roscommon.h"

#define FEATURE_MAX_D 128
#define MINPQ_INIT_NALLOCD 512

/* RANSAC error tolerance in pixels */
#define RANSAC_ERR_TOL 3

/** pessimistic estimate of fraction of inlers for RANSAC */
#define RANSAC_INLIER_FRAC_EST 0.25

/** estimate of the probability that a correspondence supports a bad model */
#define RANSAC_PROB_BAD_SUPP 0.10

/** default number of sampled intervals per octave */
#define SIFT_INTVLS 3

/** default sigma for initial gaussian smoothing */
#define SIFT_SIGMA 1.6

/** default threshold on keypoint contrast |D(x)| */
#define SIFT_CONTR_THR 0.04

/** default threshold on keypoint ratio of principle curvatures */
#define SIFT_CURV_THR 10

/** double image size before pyramid construction? */
#define SIFT_IMG_DBL 1

/** default width of descriptor histogram array */
#define SIFT_DESCR_WIDTH 4

/** default number of bins per histogram in descriptor array */
#define SIFT_DESCR_HIST_BINS 8

/* assumed gaussian blur for input image */
#define SIFT_INIT_SIGMA 0.5

/* width of border in which to ignore keypoints */
#define SIFT_IMG_BORDER 5

/* maximum steps of keypoint interpolation before failure */
#define SIFT_MAX_INTERP_STEPS 5

/* default number of bins in histogram for orientation assignment */
#define SIFT_ORI_HIST_BINS 36

/* determines gaussian sigma for orientation assignment */
#define SIFT_ORI_SIG_FCTR 1.5

/* determines the radius of the region used in orientation assignment */
#define SIFT_ORI_RADIUS 3.0 * SIFT_ORI_SIG_FCTR

/* number of passes of orientation histogram smoothing */
#define SIFT_ORI_SMOOTH_PASSES 2

/* orientation magnitude relative to max that results in new feature */
#define SIFT_ORI_PEAK_RATIO 0.8

/* determines the size of a single descriptor orientation histogram */
#define SIFT_DESCR_SCL_FCTR 3.0

/* threshold on magnitude of elements of descriptor vector */
#define SIFT_DESCR_MAG_THR 0.2

/* factor used to convert floating-point descriptor to unsigned char */
#define SIFT_INT_DESCR_FCTR 512.0

/* returns a feature's detection data */
#define feat_detection_data(f) ( (struct detection_data*)(f->feature_data) )

/* colors in which to display different feature types */
#define FEATURE_OXFD_COLOR CV_RGB(255,255,0)
#define FEATURE_LOWE_COLOR CV_RGB(255,0,255)
/* extracts a feature's RANSAC data */
#define feat_ransac_data( feat ) ( (struct ransac_data*) (feat)->feature_data )
typedef CvMat* (*ransac_xform_fn)( CvPoint2D64f* pts, CvPoint2D64f* mpts,
				   int n );
typedef double (*ransac_err_fn)( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* T );



struct lowelight{
  double x;                      /**< x coord */
  double y;                      /**< y coord */
  double scl;                    /**< scale of a Lowe-style feature */
  double ori;                    /**< orientation of a Lowe-style feature */
  unsigned char data[128];	  //data that is btw only char values
};


struct detection_data
{
  int r;
  int c;
  int octv;
  int intvl;
  double subintvl;
  double scl_octv;
};

struct feature
{
  double x;                      /**< x coord */
  double y;                      /**< y coord */
  double a;                      /**< Oxford-type affine region parameter */
  double b;                      /**< Oxford-type affine region parameter */
  double c;                      /**< Oxford-type affine region parameter */
  double scl;                    /**< scale of a Lowe-style feature */
  double ori;                    /**< orientation of a Lowe-style feature */
  int d;                         /**< descriptor length */
  double descr[FEATURE_MAX_D];   /**< descriptor */
  int type;                      /**< feature type, OXFD or LOWE */
  int category;                  /**< all-purpose feature category */
  struct feature* fwd_match;     /**< matching feature from forward image */
  struct feature* bck_match;     /**< matching feature from backmward image */
  struct feature* mdl_match;     /**< matching feature from model */
  CvPoint2D64f img_pt;           /**< location in image */
  CvPoint2D64f mdl_pt;           /**< location in model */
  void* feature_data;            /**< user-definable data */
};

enum feature_type
  {
    FEATURE_OXFD,
    FEATURE_LOWE,
  };

enum feature_match_type
  {
    FEATURE_FWD_MATCH,
    FEATURE_BCK_MATCH,
    FEATURE_MDL_MATCH,
  };

struct kd_node
{
  int ki;                      /**< partition key index */
  double kv;                   /**< partition key value */
  int leaf;                    /**< 1 if node is a leaf, 0 otherwise */
  struct feature* features;    /**< features at this node */
  int n;                       /**< number of features */
  struct kd_node* kd_left;     /**< left child */
  struct kd_node* kd_right;    /**< right child */
};

struct bbf_data
{
  double d;
  void* old_data;
};

struct ransac_data
{
  void* orig_feat_data;
  int sampled;
};

struct pq_node
{
  void* data;
  int key;
};


/** a minimizing priority queue */
struct min_pq
{
  struct pq_node* pq_array;    /* array containing priority queue */
  int nallocd;                 /* number of elements allocated */
  int n;                       /**< number of elements in pq */
};

struct coordinates
{
  gdouble x1;
  gdouble y1;
  gdouble x2;
  gdouble y2;
};

struct transmat
{
  gdouble a00;
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
};

struct uber_coordinates
{
  int imagewidth;//size of the image where we found this 
  int imageheight;//
  int mwidth;//size of the model that we used to find this this   
  int mheight;//
  gdouble scale;//Rectangular Region where this is matched into
  gdouble angle;  
  gdouble x1;//Rectangular Region where this is matched into
  gdouble y1;
  gdouble x2;
  gdouble y2;
  gdouble a00;//transformation matrix 
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
  guint64 timestamp;
  char name[30];//name of found element
  gdouble tx;
  gdouble ty; 
  gdouble tz;
  gdouble rx;
  gdouble ry; 
  gdouble rz;  
};

struct siftset{
IplImage* init_img;
IplImage* gray8;//intermediate
IplImage* gray32;
IplImage*** gauss_pyr;
IplImage*** dog_pyr;
//next important for dereferencing
int intvls;
int octvs;
int dbl_img;
int channels;
};

struct siftheap{
int numfeat;
struct feature *feat;
struct kd_node* node;
int width;
int height;
gchar name[30]; 
int sent;//0 not send 1 sent
int advertised;
ros::Publisher* rpub;
//void* rpub;
};


gdouble DoubleSwap( double f );
int _sift_features( IplImage* img, struct feature** feat, struct siftset* siftset, int intvls,
		    double sigma, double contr_thr, int curv_thr,
		    int img_dbl, int descr_width, int descr_hist_bins );
static inline struct feature* get_match( struct feature*, int );
int get_matched_features( struct feature*, int, int, struct feature*** );
int calc_min_inliers( int, int, double, double );
struct feature** draw_ransac_sample( struct feature**, int, int, gsl_rng* );
void extract_corresp_pts( struct feature**, int, int, CvPoint2D64f**,
			  CvPoint2D64f** );
int find_consensus( struct feature**, int, int, CvMat*, ransac_err_fn,
		    double, struct feature*** );
static inline void release_mem( CvPoint2D64f*, CvPoint2D64f*,
				struct feature** );
CvMat* lsq_homog( CvPoint2D64f* pts, CvPoint2D64f* mpts, int n );
double homog_xfer_err( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* H );
CvPoint2D64f persp_xform_pt( CvPoint2D64f pt, CvMat* T );

struct kd_node* kd_node_init( struct feature*, int );
void expand_kd_node_subtree( struct kd_node* );
void assign_part_key( struct kd_node* );
double median_select( double*, int );
double rank_select( double*, int, int );
void insertion_sort( double*, int );
int partition_array( double*, int, double );
void partition_features( struct kd_node* );
struct kd_node* explore_to_leaf( struct kd_node*, struct feature*,
				 struct min_pq* );
int insert_into_nbr_array( struct feature*, struct feature**, int, int );
int within_rect( CvPoint2D64f, CvRect );

void create_init_img( IplImage*, struct siftset*,int, double );
void convert_to_gray32( IplImage*, struct siftset*);
void build_pyr( struct siftset*, double );
CvSeq* scale_space_extrema( IplImage***, int, int, double, int, CvMemStorage*);
int is_extremum( IplImage***, int, int, int, int );
struct feature* interp_extremum( IplImage***, int, int, int, int, int, double);
void interp_step( IplImage***, int, int, int, int, double*, double*, double* );
CvMat* deriv_3D( IplImage***, int, int, int, int );
CvMat* hessian_3D( IplImage***, int, int, int, int );
double interp_contr( IplImage***, int, int, int, int, double, double, double );
struct feature* new_feature( void );
int is_too_edge_like( IplImage*, int, int, int );
void calc_feature_scales( CvSeq*, double, int );
void adjust_for_img_dbl( CvSeq* );
void calc_feature_oris( CvSeq*, IplImage*** );
double* ori_hist( IplImage*, int, int, int, int, double );
int calc_grad_mag_ori( IplImage*, int, int, double*, double* );
void smooth_ori_hist( double*, int );
double dominant_ori( double*, int );
void add_good_ori_features( CvSeq*, double*, int, double, struct feature* );
struct feature* clone_feature( struct feature* );
void compute_descriptors( CvSeq*, IplImage***, int, int );
double*** descr_hist( IplImage*, int, int, double, double, int, int );
void interp_hist_entry( double***, double, double, double, double, int, int);
void hist_to_descr( double***, int, int, struct feature* );
void normalize_descr( struct feature* );
int feature_cmp( void*, void*, void* );
void release_descr_hist( double****, int );
void release_pyr( IplImage****, int, int );
static inline float pixval32f( IplImage* img, int r, int c )
{
  return ( (float*)(img->imageData + img->widthStep*r) )[c];
}
void killsiftset (struct siftset* siftset);
struct kd_node* kdtree_build( struct feature* features, int n );
int kdtree_bbf_knn( struct kd_node* kd_root, struct feature* feat, int k,
		    struct feature*** nbrs, int max_nn_chks );
double descr_dist_sq( struct feature* f1, struct feature* f2 );
CvMat* ransac_xform( struct feature* features, int n, int mtype,
		     ransac_xform_fn xform_fn, int m, double p_badxform,
		     ransac_err_fn err_fn, double err_tol,
		     struct feature*** inliers, int* n_in );
void kdtree_release( struct kd_node* kd_root );

GstBuffer* gst_buf_from_oxfd(char * input);
GstBuffer* gst_buf_from_lowe(char * input);
GstBuffer* gst_buf_from_lowelight(GstBuffer * input);


struct siftset* make_siftset(IplImage* img,int intvls,int img_dbl);
