/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstrossrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION:element-rossrc
 * @see_also: #GstFileSrc
 *
 * Reads buffers from sequentially named files. If used together with an image
 * decoder, one needs to use the #Gstrossrc:caps property or a capsfilter
 * to force to caps containing a framerate. Otherwise image decoders send EOS
 * after the first picture.
 *
 * File names are created by replacing "%%d" with the index using printf().
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch rossrc location="img.%04d.png" index=0 caps="image/png,framerate=\(fraction\)12/1" ! \
 *     pngdec ! ffmpegcolorspace ! theoraenc ! oggmux ! \
 *     filesink location="images.ogg"
 * ]| This pipeline creates a video file "images.ogg" by joining multiple PNG
 * files named img.0000.png, img.0001.png, etc.
 * </refsect2>
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>

//#include <sys/inotify.h>
#include <sys/select.h>

#include <sys/types.h>
#include  <signal.h>
#include <sys/time.h>
#include <unistd.h>



#include <ros/ros.h>
#include <sensor_msgs/Image.h>
//#include <cv_bridge/CvBridge.h>
#include <image_transport/image_transport.h>

#include "gstrossrc.h"
#include "roscommon.h"


static GstFlowReturn gst_rossrc_create (GstPushSrc * src,
    GstBuffer ** buffer);

static void gst_rossrc_dispose (GObject * object);

static void gst_rossrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_rossrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps *gst_rossrc_getcaps (GstBaseSrc * src);
static gboolean gst_rossrc_query (GstBaseSrc * src, GstQuery * query);
static gboolean gst_rossrc_stop (GstBaseSrc * src);
static gboolean gst_rossrc_start (GstBaseSrc * src);
static GstStaticPadTemplate gst_rossrc_pad_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

GST_DEBUG_CATEGORY_STATIC (gst_rossrc_debug);
#define GST_CAT_DEFAULT gst_rossrc_debug

static const GstElementDetails gst_rossrc_details =
GST_ELEMENT_DETAILS ("ROS Source",
    "Source/ROS",
    "Subscribe to ROS messages on a topic",
    "Hannes Bistry <bistryATinformatikDOTuniMINUShamburgDOTde>");

enum
{
  ARG_0,
  ARG_CAPS,
  ARG_TOPIC,
  ARG_TRANSPORT
};

#define DEFAULT_LOCATION "%05d"
#define DEFAULT_INDEX 0

GST_BOILERPLATE (Gstrossrc, gst_rossrc, GstPushSrc,
    GST_TYPE_PUSH_SRC);

static void
gst_rossrc_base_init (gpointer g_class)
{
      int schrott=0;
      ros::init(schrott, NULL, "rossrc", ros::init_options::AnonymousName);

  GstElementClass *gstelement_class = GST_ELEMENT_CLASS (g_class);

  GST_DEBUG_CATEGORY_INIT (gst_rossrc_debug, "rossrc", 0,
      "rossrc element");
  
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&gst_rossrc_pad_template));
  gst_element_class_set_details (gstelement_class, &gst_rossrc_details);
}

static void
gst_rossrc_class_init (GstrossrcClass * klass)
{
  
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS (klass);
  GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS (klass);

  gobject_class->set_property = gst_rossrc_set_property;
  gobject_class->get_property = gst_rossrc_get_property;


  g_object_class_install_property (gobject_class, ARG_CAPS,
      g_param_spec_boxed ("caps", "Caps",
          "Caps describing the format of the data.",
          GST_TYPE_CAPS, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TOPIC,
      g_param_spec_string ("topic", "Topic",
          "Topic to subscribe to.",
          "/gstros/image_raw", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TRANSPORT,
      g_param_spec_string ("transport", "Transport",
          "Transport method (raw/compressed/theora/...) .",
          "raw", G_PARAM_READWRITE));



  gobject_class->dispose = gst_rossrc_dispose;

  gstbasesrc_class->get_caps = gst_rossrc_getcaps;
  gstbasesrc_class->query = gst_rossrc_query;
  gstbasesrc_class->start = gst_rossrc_start;
  gstbasesrc_class->stop = gst_rossrc_stop;
  gstpushsrc_class->create = gst_rossrc_create;



  if (sizeof (off_t) < 8) {
    GST_LOG ("No large file support, sizeof (off_t) = %" G_GSIZE_FORMAT,
        sizeof (off_t));
  }
}

static void
gst_rossrc_init (Gstrossrc * rossrc,
    GstrossrcClass * g_class)
{
  rossrc->successful_read = FALSE;
  rossrc->topic=g_strdup_printf("/gstros/image_raw");
  rossrc->transport=g_strdup_printf("raw");
  //rossrc->buffers=NULL;
  gst_base_src_set_format (GST_BASE_SRC (rossrc), GST_FORMAT_TIME);
  gst_base_src_set_live (GST_BASE_SRC (rossrc), TRUE);  

}

static void
gst_rossrc_dispose (GObject * object)
{
  Gstrossrc *src = GST_ROSSRC (object);

  g_free (src->dirname);
  src->dirname = NULL;
  if (src->newcaps)
    gst_caps_unref (src->newcaps);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static GstCaps *
gst_rossrc_getcaps (GstBaseSrc * src)
{
  Gstrossrc *multi_file_src = GST_ROSSRC (src);

  GST_DEBUG_OBJECT (src, "returning %" GST_PTR_FORMAT, multi_file_src->newcaps);

  if (multi_file_src->newcaps) {
    return gst_caps_ref (multi_file_src->newcaps);
  } else {
    return gst_caps_new_any ();
  }
}

/*Gstrossrc *globalsrc;

void image_cb(const sensor_msgs::ImageConstPtr& msg)
 {
  //boost::lock_guard<boost::mutex> guard(image_mutex_);
	   
// Hang on to message pointer for sake of mouse_cb
//last_msg_ = msg;

// May want to view raw bayer data
// NB: This is hacky, but should be OK since we have only one image CB.
//  if (msg->encoding.find("bayer") != std::string::npos) 
  g_print("Got Image Transport Message\n");


  g_mutex_lock (globalsrc->data_mutex);
  //current_data = data;
  //apply data to common buffer

  g_cond_signal (globalsrc->data_cond);
  g_mutex_unlock (globalsrc->data_mutex);
}
*/


void * spinner (void* data){
  Gstrossrc *insrc=(Gstrossrc *)data;
/*    static int spinned=0;
  if (spinned==0){
    spinned=1;
    ros::spin();
    insrc->end=TRUE;
  }
  else ros::waitForShutdown();
  //Awake my thread
    g_mutex_lock (insrc->data_mutex);
    g_cond_signal (insrc->data_cond);
    g_mutex_unlock (insrc->data_mutex);

    g_print("Returned from spinning\n");
*/
  commonspinner(insrc->data_cond,insrc->data_mutex);
  insrc->end=TRUE;
  //gst_rossrc_stop ((GstBaseSrc *) insrc);
}


static gboolean gst_rossrc_start (GstBaseSrc * src){

  Gstrossrc *insrc;

  insrc = GST_ROSSRC (src);
  insrc->end=FALSE;
  insrc->sentbuffers=NULL;
  insrc->data_cond=g_cond_new(); /* Must be initialized somewhere */
  insrc->data_mutex=g_mutex_new();
  insrc->newcaps=NULL;
  insrc->zaehler=30;
  insrc->nenner=1;
  insrc->width=-1;
  insrc->height=-1;
  insrc->silent=1;

  insrc->format=new std::string();

  //image_transport::ImageTransport it(insrc->nh);
insrc->nh = new ros::NodeHandle();
  image_transport::ImageTransport it(*insrc->nh);
  //insrc->sub = it.subscribe(std::string("/gscam/image_raw"), 1, &Gstrossrc::imageCallback, insrc, image_transport::TransportHints());
  //insrc->sub = it.subscribe(std::string("/gscam/image_raw"), 1, &Gstrossrc::imageCallback, insrc,image_transport::TransportHints("raw",ros::TransportHints(),ros::NodeHandle("~"),"image_transport"));
  insrc->sub = it.subscribe(std::string(insrc->topic), 1, &Gstrossrc::imageCallback, insrc,image_transport::TransportHints(insrc->transport,ros::TransportHints(),ros::NodeHandle("~"),"image_transport"));
pthread_create (&insrc->rosspin  , NULL, spinner, insrc);
  return TRUE;
}

static gboolean gst_rossrc_stop (GstBaseSrc * src){
  Gstrossrc *insrc;
  //g_print("Stop called\n");
  insrc = GST_ROSSRC (src);
  insrc->end=TRUE;
  insrc->sub.shutdown();
  return TRUE;
}

static gboolean
gst_rossrc_query (GstBaseSrc * src, GstQuery * query)
{
  gboolean res;
  Gstrossrc *mfsrc;
  g_print("Rossrc query now\n");
  mfsrc = GST_ROSSRC (src);

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_POSITION:
    {
      GstFormat format;

      gst_query_parse_position (query, &format, NULL);
      switch (format) {
        case GST_FORMAT_BUFFERS:
        case GST_FORMAT_DEFAULT:
          gst_query_set_position (query, GST_FORMAT_BUFFERS, mfsrc->index);
          res = TRUE;
          break;
        default:
          res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
          break;
      }
      break;
    }
    default:
      res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
      break;
  }
  return res;
}





static void
gst_rossrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstrossrc *src = GST_ROSSRC (object);

  switch (prop_id) {
    case ARG_CAPS:
    {
      const GstCaps *caps = gst_value_get_caps (value);
      GstCaps *new_caps;

      if (caps == NULL) {
        new_caps = gst_caps_new_any ();
      } else {
        new_caps = gst_caps_copy (caps);
      }
      gst_caps_replace (&src->caps, new_caps);
      gst_pad_set_caps (GST_BASE_SRC_PAD (src), new_caps);
    }
      break;
    case ARG_TOPIC:
    {
      if (src->topic)g_free(src->topic);
      src->topic=g_strdup(g_value_get_string (value)); 
    }
      break;
    case ARG_TRANSPORT:
    {
      if (src->transport)g_free(src->transport);
      src->transport=g_strdup(g_value_get_string (value)); 
    }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_rossrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstrossrc *src = GST_ROSSRC (object);

  switch (prop_id) {
    case ARG_CAPS:
      gst_value_set_caps (value, src->caps);
      break;
    case ARG_TOPIC:
      g_value_set_string (value, src->topic);
      break;
    case ARG_TRANSPORT:
      g_value_set_string (value, src->transport);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static GstFlowReturn
gst_rossrc_create (GstPushSrc * src, GstBuffer ** buffer)
{
  Gstrossrc *insrc;
  gsize size;
  gchar *data;
  GstBuffer *buf;
  gboolean ret;
  GError *error = NULL;

  insrc = GST_ROSSRC (src);
  int result;
  if (insrc->end){
        g_print("Will now return Unexpected\n");
	return GST_FLOW_UNEXPECTED; 
  }
  if (!insrc->silent)g_print("Will lock mutex now\n");
  g_mutex_lock (insrc->data_mutex);
  //kicking old buffers
  GList* temp=insrc->sentbuffers;
  while(temp){
    if (((GstMiniObject*)(temp->data))->refcount==1){
	//search for Data pointer
	for (std::vector<sensor_msgs::ImageConstPtr>::iterator it = insrc->msglist.begin() ; it != insrc->msglist.end(); ++it){
	  //g_print("Considering to delete %p\n",(&((*it)->data[0])));
	  if ((&((*it)->data[0]))==((GstBuffer*)(temp->data))->data) {
		insrc->msglist.erase(it);
		break;
	  }
	}

	gst_buffer_unref((GstBuffer*)temp->data);
	insrc->sentbuffers=g_list_delete_link(insrc->sentbuffers,temp);
	temp=insrc->sentbuffers;
    }	
    else temp=g_list_next(temp);
  }

  //while (!insrc->buffer)
    g_cond_wait (insrc->data_cond, insrc->data_mutex);
  //data = current_data;
  //current_data = NULL;
  if (!insrc->silent)g_print("Awaked from waiting for buffer\n");

  
  if (!insrc->current){
        if (!insrc->silent)g_print("Awaked, but no buffer present\n");
        g_mutex_unlock (insrc->data_mutex);
	return GST_FLOW_UNEXPECTED; 
  }


  insrc->successful_read = TRUE;


  //GST_DEBUG_OBJECT (insrc, "read file \"%s\".", filename);

  *buffer = gst_buffer_ref(insrc->current);
  insrc->sentbuffers=g_list_append(insrc->sentbuffers,insrc->current);
  insrc->current=NULL;
  g_mutex_unlock (insrc->data_mutex);
  return GST_FLOW_OK;
}




//static gboolean init_rossrc (GstPlugin * plugin){
//return gst_element_register (plugin, "rossrc", GST_RANK_NONE, GST_TYPE_ROSSRC);}
//GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,GST_VERSION_MINOR,"rossrc","Template rossrc",init_rossrc,VERSION,"LGPL","GStreamer","http:gstreamer.net")
